package com.blazinggames.ootw.ep32
{
	// Adobe Imports
	import flash.display.Graphics;
	import flash.display.MovieClip;
	import flash.display.Sprite;

	public class Forest
	extends MovieClip
	{
// *********************
// ***** CONSTANTS *****
// *********************
		public const TILES:String = "s-x";
		public const TILE_START = 0;
		public const TILE_EMPTY = 1;
		public const TILE_TREE = 2;
		public const TILE_UNKNOWN = 3;

		public const RAW_MAP:String = 
"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"+
"x------------------------------x"+
"x------------------------------x"+
"x-x--x--x--x--x--x--x--x--x--x-x"+
"x------------------------------x"+
"x------------------------------x"+
"x--x--x--x--x--x--x--x--x--x--xx"+
"x------------------------------x"+
"x------------------------------x"+
"x-x--x--x--x--x--x--x--x--x--x-x"+
"x------------------------------x"+
"x------------------------------x"+
"x--x--x--x--x--x--x--x--x--x--xx"+
"x------------------------------x"+
"x------------------------------x"+
"x-x--x--x--x--x--x--x--x--x--x-x"+
"x------------------------------x"+
"x------------------------------x"+
"x--x--x--x--x--x--x--x--x--x--xx"+
"x------------------------------x"+
"x------------------------------x"+
"x-x--x--x--x--x--x--x--x--x--x-x"+
"x------------------------------x"+
"x------------------------------x"+
"x--x--x--x--x--x--x--x--x--x--xx"+
"x------------------------------x"+
"x------------------------------x"+
"x-x--x--x--x--x--x--x--x--x--x-x"+
"x------------------------------x"+
"x------------------------------x"+
"x--x--x--x--x--x--x--x--x--x--xx"+
"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";

		protected const AUTOMAPCOLORS:Array = new Array(
			0xffff00, 0xffffff, 0x00ff00, 0x808080
		);

// *********************
// ***** VARIABLES *****
// *********************

		// for map management
		private var _map:Array;
		private var _automap:Array;
		private var _sprites:Array;
		private var _forest:MovieClip;
		private var _automapDisplay:Sprite;
		private var _playerPositions:Sprite;
		
		// character positions
		private var _ykcul:Ykcul;
		private var _squirrel:Squirrel;
		private var _player:Player;
		private var _curCharacter:int;
		private var _playersSprites:Array;
		private var _playersD:Array;
		private var _needAutomapUpdate:Boolean;
		private var _charles:Charles;

		// time tracking
		private var _playTime:Number;

// ***********************
// ***** CONSTRUCTOR *****
// ***********************

		public function Forest()
		{
			var cntr:int;
			var tempTree:Tree;
			var bd:BillboardData;

			trace ("Starting initialization");
			// build the map
			_map = new Array(1024);
			_automap = new Array(1024);
			_sprites = new Array();
			_forest = new MovieClip();
			for (cntr = 0; cntr < 1024; ++cntr) {
				_map[cntr] = TILES.indexOf(RAW_MAP.charAt(cntr));
				if (_map[cntr] == TILE_TREE) {
					tempTree = new Tree(cntr % 32+.5, cntr / 32+.5);
					bd = tempTree.getBillboardData();
					_forest.addChild(tempTree);
					_sprites.push(tempTree);
				}
			}
			addChild(_forest);
			_automapDisplay = new Sprite();
			addChild(_automapDisplay);
			_playerPositions = new Sprite();
			addChild(_playerPositions);

			_curCharacter = 0;
			_ykcul = new Ykcul(16.0, 29.5);
			_forest.addChild(_ykcul);
			_sprites.push(_ykcul);
			_squirrel = new Squirrel(17, 29.5);
			_forest.addChild(_squirrel);
			_sprites.push(_squirrel);
			_player = new Player(15, 29.5);
			_forest.addChild(_player);
			_sprites.push(_player);
			_playersSprites = new Array(_player, _ykcul, _squirrel);
			_playersD = new Array(0,0,0);
			
			_charles = new Charles(16.5,16.5);
			_forest.addChild(_charles);
			_sprites.push(_charles);
			resetMap(true);
			render();
			trace ("Ending initialization");
		}

// *****************************************
// ***** Character movement management *****
// *****************************************

		public function moveForward(dist:Number):int
		{
			var angle:Number = _playersD[_curCharacter] * Math.PI/1600;
			var billboard = _playersSprites[_curCharacter].getBillboardData();
			var nx:Number = billboard.mapX + dist * Math.sin(angle);
			var ny:Number = billboard.mapY - dist * Math.cos(angle);
			var nt:int = getTile(uint(nx), uint(ny));
			if (nt < TILE_TREE) {
				_needAutomapUpdate = 
						( (Math.floor(nx) != Math.floor(billboard.mapX)) ||
						(Math.floor(ny) != Math.floor(billboard.mapY)));
				billboard.mapX = nx;
				billboard.mapY = ny;
			}
			return 0;
		}
		
		public function moveBackward(dist:Number)
		{
			_playersD[_curCharacter] -= 1600; // 180
			moveForward(dist);
			_playersD[_curCharacter] += 1600; // 180
		}
		
		public function turnLeft(dist:Number)
		{
			_playersD[_curCharacter] -= (dist*800);
		}
		
		public function turnRight(dist:Number)
		{
			_playersD[_curCharacter] += (dist*800);
		}

		public function setCurrentCharacter(n:uint)
		{
			if (n > 2) return;
			_curCharacter = n;
		}
		
		public function getCurrentCharacter():int
		{
			return _curCharacter;
		}
		
		/**
		* @param cc the character who's coordinates are desired. 
		* Default is current character.
		* @return the x coordinate of the current or indicated character
		*/
		public function getPlayerX(cc:int = -1):Number
		{
			var temp:int = cc >= 0 ? cc : _curCharacter;
			var billboard = _playersSprites[temp].getBillboardData();
			
			return billboard.mapX;
		}

		/**
		* @param cc the character who's coordinates are desired. 
		* Default is current character.
		* @return the y coordinate of the current or indicated character
		*/
		public function getPlayerY(cc:int = -1):Number
		{
			var temp:int = cc >= 0 ? cc : _curCharacter;
			var billboard = _playersSprites[temp].getBillboardData();
			
			return billboard.mapY;
		}
		
		/**
		* @param cc the character who's coordinates are desired. 
		* Default is current character.
		* @return the dir coordinate of the current or indicated character
		*/
		public function getPlayerDirection(cc:int = -1):Number
		{
			var temp:int = cc >= 0 ? cc : _curCharacter;
			
			return _playersD[temp];
		}

		public function setSex(sex:int)
		{
			_player.setSex(sex);
		}

// *****************************
// ***** Trigonometry aids *****
// *****************************

		public function getAngle(dx:Number, dy:Number, uic:int):Number
		{
			
			var ang:Number = Math.atan2(dy, dx) * uic/2/Math.PI + uic/4;
			if (ang < 0) 
				ang += uic;
			else if (ang > uic)
				ang -= uic;
			return ang;
		}

		public function getDistance(x1:Number, y1:Number, 
				x2:Number, y2:Number):Number
		{
			var dx:Number = x1 - x2;
			var dy:Number = y1 - y2;
			return Math.sqrt(dx*dx + dy*dy + .00000001);
		}

// **************************************
// ***** Functions - Map Management *****
// **************************************

		public function getTile(x:uint, y:uint):int
		{
			var mapX = x & 31;
			var mapY = y & 31;
			return _map[(mapY*32) + mapX];
		}

		public function resetMap(clearAuto:Boolean = false)
		{
			_playTime = 0;
			_ykcul.getBillboardData().setMapLocation(16.0, 29.5);
			_squirrel.getBillboardData().setMapLocation(17, 29.5);
			_player.getBillboardData().setMapLocation(15, 29.5);
			_playersD = new Array(0,0,0);
			_charles.getBillboardData().setMapLocation(16.5,16.5);
			if (clearAuto) {
				for (var cntr:int = 0; cntr < 1024; ++cntr)
					_automap[cntr] = TILE_UNKNOWN;
			}
			_needAutomapUpdate = true;
			updateAutomap();
		}

		public function generateAutomap()
		{
			var tileIndex:int = 0;
			var g:Graphics = _automapDisplay.graphics;
			g.clear();
			for (var cntrY:int = 0; cntrY < 32; ++cntrY) {
				for (var cntrX:int = 0; cntrX < 32; ++cntrX) {
					g.beginFill(AUTOMAPCOLORS[_map[tileIndex]]);
					g.drawRect(cntrX*4, cntrY*4, 4, 4);
					g.endFill();
					++tileIndex;
				}
			}

			_needAutomapUpdate = false;
		}

		public function updateAutomap()
		{
			if (_needAutomapUpdate)
				generateAutomap();

			var g:Graphics = _playerPositions.graphics;
			g.clear();
			var angle:Number = _playersD[_curCharacter] * Math.PI/1600;
			var ib:IBillboard;
			var bbd:BillboardData;
			for (var cntrPlayer:int = 0; cntrPlayer < 3; ++cntrPlayer) {
				ib = IBillboard(_playersSprites[cntrPlayer]);
				bbd = ib.getBillboardData();
				g.beginFill(0xff00ff);
				g.drawRect(bbd.mapX * 4 - 2, bbd.mapY * 4 -2, 5, 5);
				g.endFill();
			}
			bbd = _charles.getBillboardData();
			g.beginFill(0xff0000);
			g.drawRect(bbd.mapX * 4 - 2, bbd.mapY * 4 -2, 5, 5);
			g.endFill();
			bbd = _playersSprites[_curCharacter].getBillboardData();
			var nx:Number = bbd.mapX + 4 * Math.sin(angle);
			var ny:Number = bbd.mapY - 4 * Math.cos(angle);
			g.lineStyle(1, 0xff00ff);
			g.moveTo(bbd.mapX * 4, bbd.mapY * 4);
			g.lineTo(nx*4, ny*4);
		}
		
		public function render(secs:Number = 1):Boolean
		{
			var cntr:int, cntrZ:int;
			var ang:Number;
			var zBillboard:IBillboard;
			var tbd:BillboardData = _playersSprites[_curCharacter].getBillboardData();
			var px:Number = tbd.mapX;
			var py:Number = tbd.mapY;
			var ts:MovieClip;
			var hght:Number;
			var wdth:Number;
			var visibleTrees:int = 0;
			var indx:int;
			var dist:Number;
			var ground:Number;
			
			// first, move charles
			_charles.takeTurn(this, secs);
			_playTime += secs;

			// loop through the trees and render the visible ones
			for (cntr = 0; cntr < _sprites.length; ++ cntr) {
				ts = _sprites[cntr];
				zBillboard = IBillboard(_sprites[cntr]);
				tbd = zBillboard.getBillboardData();
				ang = getAngle(tbd.mapX - px, tbd.mapY - py, 3200);
				ang -= _playersD[_curCharacter];
				ang += 640;//320;
				while (ang < 0) ang += 3200;
				while (ang > 3200) ang -= 3200;
				if (ang < 1280) {
					ts.visible = true;

					dist = getDistance(tbd.mapX, tbd.mapY, px, py);
					ground = 240 + 500.0 / dist;
					hght = tbd.nearHeight / dist;
					wdth = tbd.nearWidth / dist;
					ts.x = ang - 320 - wdth/2;
					ts.y = ground - hght;
					ts.width = wdth;
					ts.height = hght;
					// z sort the tree against other billboards
					indx = visibleTrees;
					for (cntrZ = 0; cntrZ < visibleTrees; ++cntrZ) {
						zBillboard = IBillboard(_forest.getChildAt(cntrZ));
						if (dist > zBillboard.getBillboardData().getDistance()) {
							indx = cntrZ;
							break;
						}
					}
					_forest.setChildIndex(ts, indx);
					tbd.setDistance(dist);
					++visibleTrees;

				} else
					ts.visible = false;
				_playersSprites[_curCharacter].visible = false;
//				trace (ts.mapX, ts.mapY, ang);
//				ts.visible = ang < 640;
			}
			updateAutomap();
			var distanceFromCharles:Number = 999;
			tbd = _charles.getBillboardData();
			for (cntr = 0; cntr < 3; ++cntr) {
				dist = getDistance(
						getPlayerX(cntr),getPlayerY(cntr), tbd.mapX,tbd.mapY);
				if (distanceFromCharles > dist) 
					distanceFromCharles = dist;
			}

			return (distanceFromCharles < 1.0);
		}

		public function getPlayTime():Number
		{
			return _playTime;
		}
	} // END CLASS
} // END PACKAGE
